# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from typing import Callable
from abc import abstractmethod

from .base import install_uvloop, Utils
from .socket import UnixSocketServer, UnixSocketClient, DEFAULT_LIMIT

from ..interface import RunnableInterface
from ..process import Daemon


class MainProcessAbstract(Daemon):

    def __init__(
            self, target: Callable, sub_process_num: int, *,
            set_affinity=None, join_timeout: int = 10,
            unix_socket_path=r'/tmp/unix_socket_endpoint', unix_socket_limit=DEFAULT_LIMIT,
            **kwargs
    ):

        super().__init__(
            target, sub_process_num, set_affinity=set_affinity, join_timeout=join_timeout,
            unix_socket_path=unix_socket_path,
            **kwargs
        )

        self._socket_server = UnixSocketServer(self._client_connected_cb, unix_socket_path, unix_socket_limit)

    @abstractmethod
    async def _client_connected_cb(self, reader, writer):
        raise NotImplementedError()

    async def _run(self):

        self._fill_process()

        while self._check_process():
            await Utils.sleep(1)

    def run(self):

        Utils.print_slogan()

        install_uvloop()

        Utils.run_until_complete(self._socket_server.open())
        Utils.run_until_complete(self._run())
        Utils.run_until_complete(self._socket_server.close())


class SubProcessAbstract(RunnableInterface):

    @classmethod
    def create(cls, process_num, unix_socket_path):

        cls(process_num, unix_socket_path).run()

    def __init__(self, process_num, unix_socket_path: str):

        self._socket_client = UnixSocketClient(unix_socket_path)

        self._process_id = Utils.getpid()
        self._process_num = process_num

    @abstractmethod
    async def _run(self):
        raise NotImplementedError()

    def run(self):

        Utils.log.success(f'Started worker process [id:{self._process_id} num:{self._process_num}]')

        install_uvloop()

        Utils.run_until_complete(self._socket_client.open())
        Utils.run_until_complete(self._run())
        Utils.run_until_complete(self._socket_client.close())

        Utils.log.success(f'Stopped worker process [id:{self._process_id} num:{self._process_num}]')



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import typing
import signal
import socket
import asyncio
import base64
import msgpack

from typing import List

from ...extend.asyncio.base import Utils, install_uvloop
from ...extend.interface import RunnableInterface

from ..error import RouterError


DEFAULT_LIMIT = 0xffffff


async def recv_msg(reader, timeout=None):

    result = None

    with Utils.suppress(asyncio.TimeoutError):

        data = await asyncio.wait_for(reader.readline(), timeout)

        if data:
            result = msgpack.loads(
                base64.b64decode(data)
            )

    return result


async def send_msg(writer, data):

    writer.writelines(
        [
            base64.b64encode(
                msgpack.dumps(data)
            ),
            b'\n',
        ]
    )

    await writer.drain()


class UnixSocketServer:

    def __init__(self, client_connected_cb, path, limit=DEFAULT_LIMIT):

        self._client_connected_cb = client_connected_cb

        self._path = path
        self._limit = limit

        self._server = None

    async def open(self):

        if os.path.exists(self._path):
            os.remove(self._path)

        self._server = await asyncio.start_unix_server(
            self._client_connected_cb, self._path, limit=self._limit
        )

        await self._server.start_serving()

    async def close(self):

        if self._server is not None:
            self._server.close()
            self._server = None


class UnixSocketClient:

    def __init__(self, path, limit=DEFAULT_LIMIT):

        self._path = path
        self._limit = limit

        self._reader = None
        self._writer = None

    @property
    def reader(self):
        return self._reader

    @property
    def writer(self):
        return self._writer

    async def open(self):

        self._reader, self._writer = await asyncio.open_unix_connection(
            self._path,
            limit=self._limit
        )

    async def close(self):

        self._reader = None

        if self._writer is not None:
            self._writer.close()
            await self._writer.wait_closed()
            self._writer = None

    async def recv_msg(self, timeout=None):

        return await recv_msg(self._reader, timeout)

    async def send_msg(self, data):

        return await send_msg(self._writer, data)


class Router:

    def __init__(self, root=r''):

        self._root = root
        self._reg_func = {}

    def __repr__(self):

        return self._reg_func.__repr__()

    def __call__(self, method, *args, **kwargs):

        func = self._reg_func.get(method)

        if not func:
            raise RouterError(f'{method} not exists')

        return func(*args, **kwargs)

    def _reg(self, method, func):

        _method = f'{self._root}{method}'

        if _method in self._reg_func:
            raise RouterError(f'{method} has exists')

        self._reg_func[_method] = func

    def reg(self, method):

        def _reg_func(func):
            self._reg(method, func)
            return func

        return _reg_func

    def items(self):

        return self._reg_func.items()

    def include(self, router):

        for method, func in router.items():
            self._reg(method, func)


class SocketConfig:

    __slots__ = [r'client_connected_cb', r'address', r'family', r'backlog', r'reuse_port', r'buffer_limit']

    def __init__(self, client_connected_cb, address, family, backlog, reuse_port, buffer_limit):

        self.client_connected_cb = client_connected_cb
        self.address = address
        self.family = family
        self.backlog = backlog
        self.reuse_port = reuse_port
        self.buffer_limit = buffer_limit


class AsyncTcpServer(RunnableInterface):

    def __init__(
            self, client_connected_cb, address, *,
            family=socket.AF_INET, backlog=None, reuse_port=True, buffer_limit=DEFAULT_LIMIT,
            on_startup=None, on_shutdown=None
    ):

        self._listeners: List[SocketConfig] = [
            SocketConfig(
                client_connected_cb, address,
                family, backlog, reuse_port, buffer_limit
            )
        ]

        self._on_startup = on_startup
        self._on_shutdown = on_shutdown

        self._servers: typing.List[asyncio.AbstractServer] = []

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

    async def __aenter__(self):

        for server in self._servers:
            await server.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, _traceback):

        for server in self._servers:
            await server.__aexit__(exc_type, exc_value, _traceback)

    def add_listener(
            self, client_connected_cb, address, *,
            family=socket.AF_INET, backlog=None, reuse_port=True, buffer_limit=DEFAULT_LIMIT
    ):

        self._listeners.append(
            SocketConfig(
                client_connected_cb, address,
                family, backlog, reuse_port, buffer_limit
            )
        )

    def run(self, *, debug=None):

        Utils.print_slogan()

        install_uvloop()

        asyncio.run(self._run(), debug=debug)

    async def _run(self):

        if self._on_startup is not None:
            await self._on_startup()

        for config in self._listeners:

            if config.family == socket.AF_UNIX:

                _socket_server = await asyncio.start_unix_server(
                    config.client_connected_cb, config.address, limit=config.buffer_limit
                )

                Utils.log.success(f'unix server [pid:{Utils.getpid()}] startup complete: {config.address}')

            else:

                _socket = socket.create_server(
                    config.address, family=config.family, backlog=config.backlog, reuse_port=config.reuse_port
                )

                _socket_server = await asyncio.start_server(
                    config.client_connected_cb,
                    limit=config.buffer_limit, sock=_socket
                )

                Utils.log.success(f'socket server [pid:{Utils.getpid()}] startup complete: {config.address}')

            self._servers.append(_socket_server)

        async with self:
            await asyncio.gather(*(_server.wait_closed() for _server in self._servers))

        if self._on_shutdown is not None:
            await self._on_shutdown()

    def _exit(self, *_):

        for _server in self._servers:
            _server.close()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from .base import Utils, AsyncCirculatoryForSecond, FuncWrapper, FutureWithTimeout

from ..event import EventDispatcher as _EventDispatcher
from ..error import catch_error


class EventDispatcher(_EventDispatcher):
    """支持异步函数的事件总线
    """

    def _gen_observer(self):

        return FuncWrapper()


class DistributedEvent(EventDispatcher):
    """Redis实现的消息广播总线
    """

    def __init__(self, redis_client, channel_name, channel_count):

        super().__init__()

        self._redis_client = redis_client

        self._channels = [f'event_bus_{Utils.md5_u32(channel_name)}_{index}' for index in range(channel_count)]

        for channel in self._channels:
            Utils.create_task(self._event_listener(channel))

    async def _event_listener(self, channel):

        async for _ in AsyncCirculatoryForSecond():

            with catch_error():

                receiver = self._redis_client.pubsub(ignore_subscribe_messages=True)

                try:

                    await receiver.subscribe(channel)

                    Utils.log.info(f'event bus channel({channel}) receiver created')

                    while True:

                        message = await receiver.get_message(timeout=10)

                        if message:
                            await self._event_assigner(message)

                except Exception as err:

                    Utils.log.warning(str(err))

                finally:

                    await receiver.close()

    async def _event_assigner(self, message):

        data = Utils.msgpack_decode(message[r'data'])

        _type = data.get(r'type', r'')
        args = data.get(r'args', [])
        kwargs = data.get(r'kwargs', {})

        if _type in self._observers:
            self._observers[_type](*args, **kwargs)

    def _gen_observer(self):

        return FuncWrapper()

    async def dispatch(self, _type, *args, **kwargs):

        channel = self._channels[Utils.md5_u32(_type) % len(self._channels)]

        message = {
            r'type': _type,
            r'args': args,
            r'kwargs': kwargs,
        }

        await self._redis_client.publish(channel, Utils.msgpack_encode(message))

    def gen_event_waiter(self, event_type, delay_time):

        return EventWaiter(self, event_type, delay_time)


class EventWaiter(FutureWithTimeout):
    """带超时的临时消息接收器
    """

    def __init__(self, dispatcher, event_type, delay_time):

        super().__init__(delay_time)

        self._dispatcher = dispatcher
        self._event_type = event_type

        self._dispatcher.add_listener(self._event_type, self._event_handler)

    def set_result(self, result):

        if self.done():
            return

        super().set_result(result)

        self._dispatcher.remove_listener(self._event_type, self._event_handler)

    def _event_handler(self, *args, **kwargs):

        if not self.done():
            self.set_result({r'args': args, r'kwargs': kwargs})



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import re
import sys
import platform
import uuid
import time
import math
import string
import random
import hashlib
import base64
import hmac
import urllib
import textwrap
import itertools
import copy
import pytz
import pickle
import psutil
import functools
import binascii
import ujson
import msgpack
import zlib
import socket
import struct
import unicodedata
import calendar
import jwt
import yaml
import loguru
import warnings
import xmltodict
import xml.dom.minidom

from datetime import datetime, timedelta
from contextlib import closing, suppress
from collections import OrderedDict
from collections.abc import Iterable
from cachetools import cached, TTLCache
from zipfile import ZipFile, ZIP_DEFLATED

from stdnum import luhn
from dateutil.parser import parse as date_parse

from .text import StrUtils
from .. import hagworm_slogan


class Utils:
    """基础工具类

    集成常用的工具函数

    """

    _BYTES_TYPES = (bytes, type(None),)

    _STRING_TYPES = (str, type(None),)

    _FALSE_VALUES = (r'null', r'none', r'nil', r'false', r'0', r'', False, 0,)

    math = math
    random = random
    textwrap = textwrap
    itertools = itertools

    path = os.path

    log = loguru.logger
    urlparse = urllib.parse

    deepcopy = staticmethod(copy.deepcopy)

    func_wraps = staticmethod(functools.wraps)
    func_partial = staticmethod(functools.partial)

    randint = staticmethod(random.randint)
    randstr = staticmethod(random.sample)

    getenv = staticmethod(os.getenv)
    getpid = staticmethod(os.getpid)
    getppid = staticmethod(os.getppid)

    date_parse = staticmethod(date_parse)

    suppress = staticmethod(suppress)

    @classmethod
    def get_class_name(cls, val):

        return f'{val.__module__}.{val.__name__}'

    @classmethod
    def environment(cls):

        return {
            r'python': sys.version,
            r'system': [platform.system(), platform.release(), platform.version(), platform.machine()],
        }

    @classmethod
    def print_slogan(cls):

        environment = Utils.environment()

        Utils.log.info(
            f'{hagworm_slogan}'
            f'python {environment["python"]}\n'
            f'system {" ".join(environment["system"])}'
        )

    @classmethod
    def deprecation_warning(cls, val):

        warnings.warn(val, DeprecationWarning)

    @classmethod
    def get_node_id(cls):

        return cls.md5(f'{cls.getpid()}@{uuid.getnode()}')

    @classmethod
    def utf8(cls, val):

        if isinstance(val, cls._BYTES_TYPES):
            return val

        if isinstance(val, str):
            return val.encode(r'utf-8')

        raise TypeError(
            r'Expected str, bytes, or None; got %r' % type(val)
        )

    @classmethod
    def basestring(cls, val):

        if isinstance(val, cls._STRING_TYPES):
            return val

        if isinstance(val, bytes):
            return val.decode(r'utf-8')

        raise TypeError(
            r'Expected str, bytes, or None; got %r' % type(val)
        )

    @classmethod
    def vague_string(cls, val, vague_str=r'*', vague_len=0.5, min_vague=1, max_vague=10):

        val_len = len(val)
        vague_len = max(min_vague, min(max_vague, cls.math.ceil(val_len * vague_len)))

        r_show = (val_len - vague_len) // 2
        l_show = max(1, val_len - r_show - vague_len) if val_len > min_vague else 0

        return val[:l_show] + vague_str * vague_len + val[val_len - r_show:]

    @classmethod
    def json_encode(cls, val, **kwargs):

        return ujson.dumps(val, **kwargs)

    @classmethod
    def json_decode(cls, val, **kwargs):

        return ujson.loads(cls.basestring(val), **kwargs)

    @classmethod
    def msgpack_encode(cls, val, **kwargs):

        return msgpack.dumps(val, **kwargs)

    @classmethod
    def msgpack_decode(cls, val, **kwargs):

        return msgpack.loads(val, **kwargs)

    @classmethod
    def today(cls):

        result = datetime.now()

        return result.replace(hour=0, minute=0, second=0, microsecond=0)

    @classmethod
    def yesterday(cls):

        result = datetime.now() - timedelta(days=1)

        return result.replace(hour=0, minute=0, second=0, microsecond=0)

    @classmethod
    def localnow(cls):

        result = datetime.now()

        return result.replace(second=0, microsecond=0)

    @classmethod
    def localnow_on_hour(cls):

        result = datetime.now()

        return result.replace(minute=0, second=0, microsecond=0)

    @classmethod
    def utcnow(cls):

        result = datetime.utcnow()

        return result.replace(second=0, microsecond=0)

    @classmethod
    def utcnow_on_hour(cls):

        result = datetime.utcnow()

        return result.replace(minute=0, second=0, microsecond=0)

    @classmethod
    def timestamp(cls, msec=False):

        if msec:
            return int(time.time() * 1000)
        else:
            return int(time.time())

    @classmethod
    def convert_bool(cls, val):

        if isinstance(val, str):
            val = val.lower()

        return val not in cls._FALSE_VALUES

    @classmethod
    def convert_int(cls, val, default=0):

        result = default

        try:
            if not isinstance(val, float):
                result = int(val)
        except Exception as _:
            pass

        return result

    @classmethod
    def convert_float(cls, val, default=0.0):

        result = default

        try:
            if not isinstance(val, float):
                result = float(val)
        except Exception as _:
            pass

        return result

    @classmethod
    def interval_limit(cls, val, min_val, max_val):

        result = val

        if min_val is not None:
            result = max(result, min_val)

        if max_val is not None:
            result = min(result, max_val)

        return result

    @classmethod
    def split_float(cls, val, sep=r',', minsplit=0, maxsplit=-1):

        result = [
            float(item.strip()) for item in val.split(sep, maxsplit)
            if cls.re_match(r'[\d\.]+', item.strip().lstrip(r'-'))
        ]

        fill = minsplit - len(result)

        if fill > 0:
            result.extend(0 for _ in range(fill))

        return result

    @classmethod
    def join_float(cls, iterable, sep=r','):

        result = []

        for item in iterable:

            _type = type(item)

            if _type is float:

                result.append(str(item))

            elif _type is str:

                item = item.strip()

                if cls.re_match(r'[\d\.]+', item):
                    result.append(item)

        return sep.join(result)

    @classmethod
    def split_int(cls, val, sep=r',', minsplit=0, maxsplit=-1):

        result = [int(item.strip()) for item in val.split(sep, maxsplit) if item.strip().lstrip(r'-').isdigit()]

        fill = minsplit - len(result)

        if fill > 0:
            result.extend(0 for _ in range(fill))

        return result

    @classmethod
    def join_int(cls, iterable, sep=r','):

        result = []

        for item in iterable:

            _type = type(item)

            if _type is int:

                result.append(str(item))

            elif _type is str:

                item = item.strip()

                if item.isdigit():
                    result.append(item)

        return sep.join(result)

    @classmethod
    def split_str(cls, val, sep=r'|', minsplit=0, maxsplit=-1):

        if val:
            result = [item.strip() for item in val.split(sep, maxsplit)]
        else:
            result = []

        fill = minsplit - len(result)

        if fill > 0:
            result.extend(r'' for _ in range(fill))

        return result

    @classmethod
    def join_str(cls, iterable, sep=r'|'):

        return sep.join(str(val).replace(sep, r'') for val in iterable)

    @classmethod
    def list_extend(cls, iterable, val):

        if cls.is_iterable(val, True):
            iterable.extend(val)
        else:
            iterable.append(val)

    @classmethod
    def str_len(cls, str_val):

        result = 0

        for val in str_val:

            if unicodedata.east_asian_width(val) in (r'A', r'F', r'W'):
                result += 2
            else:
                result += 1

        return result

    @classmethod
    def sub_str(cls, str_val, length, suffix=r'...'):

        result = []
        strlen = 0

        for val in str_val:

            if unicodedata.east_asian_width(val) in (r'A', r'F', r'W'):
                strlen += 2
            else:
                strlen += 1

            if strlen > length:

                if suffix:
                    result.append(suffix)

                break

            result.append(val)

        return r''.join(result)

    @classmethod
    def re_match(cls, pattern, value):

        result = re.match(pattern, value)

        return True if result else False

    @classmethod
    def randhit(cls, iterable, prob):

        if callable(prob):
            prob = [prob(val) for val in iterable]

        prob_sum = sum(prob)

        if prob_sum > 0:

            prob_hit = 0
            prob_sum = cls.randint(0, prob_sum)

            for index in range(0, len(prob)):

                prob_hit += prob[index]

                if prob_hit >= prob_sum:
                    return iterable[index]

        return cls.random.choice(iterable)

    @classmethod
    def urandom_seed(cls):

        seed_num = None

        try:

            with open(r'/dev/urandom', r'rb') as f:
                seed_num = struct.unpack(f'!I', f.read(4))[0]
                random.seed(seed_num)

        except Exception as _:

            cls.log.warning(r'SYSTEM DOES NOT SUPPORT URANDOM')

        return seed_num

    @classmethod
    def read_urandom(cls, size):

        result = None

        try:

            with open(r'/dev/urandom', r'rb') as f:
                result = f.read(size)

        except Exception as _:

            cls.log.warning(r'SYSTEM DOES NOT SUPPORT URANDOM')

        return result

    @classmethod
    def get_host_ip(cls):

        result = None

        with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as _socket:
            _socket.connect((r'223.5.5.5', 53))
            result = _socket.getsockname()[0]

        return result

    @classmethod
    def ip2int(cls, val):

        try:
            return int(binascii.hexlify(socket.inet_aton(val)), 16)
        except socket.error:
            return int(binascii.hexlify(socket.inet_pton(socket.AF_INET6, val)), 16)

    @classmethod
    def int2ip(cls, val):

        try:
            return socket.inet_ntoa(binascii.unhexlify(r'%08x' % val))
        except socket.error:
            return socket.inet_ntop(socket.AF_INET6, binascii.unhexlify(r'%032x' % val))

    @classmethod
    def time2stamp(cls, time_str, format_type=r'%Y-%m-%d %H:%M:%S', timezone=None):

        if timezone is None:
            return int(datetime.strptime(time_str, format_type).timestamp())
        else:
            return int(datetime.strptime(time_str, format_type).replace(tzinfo=pytz.timezone(timezone)).timestamp())

    @classmethod
    def stamp2time(cls, time_int=None, format_type=r'%Y-%m-%d %H:%M:%S', timezone=None):

        if time_int is None:
            time_int = cls.timestamp()

        if timezone is None:
            return time.strftime(format_type, datetime.fromtimestamp(time_int).timetuple())
        else:
            return time.strftime(format_type, datetime.fromtimestamp(time_int, pytz.timezone(timezone)).timetuple())

    @classmethod
    def radix24(cls, val, align=0):

        base = StrUtils.SAFE_STRING_BASE

        return cls.radix_n(val, base, 24, align)

    @classmethod
    def radix24_to_10(cls, val):

        base = StrUtils.SAFE_STRING_BASE

        return cls.radix_n_to_10(val, base, 24)

    @classmethod
    def radix36(cls, val, align=0):

        base = string.digits + string.ascii_uppercase

        return cls.radix_n(val, base, 36, align)

    @classmethod
    def radix36_to_10(cls, val):

        base = string.digits + string.ascii_uppercase

        return cls.radix_n_to_10(val, base, 36)

    @classmethod
    def radix62(cls, val, align=0):

        base = string.digits + string.ascii_letters

        return cls.radix_n(val, base, 62, align)

    @classmethod
    def radix62_to_10(cls, val):

        base = string.digits + string.ascii_letters

        return cls.radix_n_to_10(val, base, 62)

    @classmethod
    def radix_n(cls, val, base, radix, align=0):

        num = abs(int(val))

        result = ''

        while num > 0:
            num, rem = divmod(num, radix)
            result = base[rem] + result

        return r'{0:0>{1:d}s}'.format(result, align)

    @classmethod
    def radix_n_to_10(cls, val, base, radix):

        result = 0

        for _str in val.strip():
            rem = base.index(_str)
            result = result * radix + rem

        return result

    @classmethod
    def xml_encode(cls, dict_val, root_tag=r'root'):

        xml_doc = xml.dom.minidom.Document()

        root_node = xml_doc.createElement(root_tag)
        xml_doc.appendChild(root_node)

        def _convert(_doc, _node, _dict):

            for key, val in _dict.items():

                temp = _doc.createElement(key)

                if isinstance(val, dict):
                    _convert(_doc, temp, val)
                else:
                    temp.appendChild(_doc.createTextNode(str(val)))

                _node.appendChild(temp)

        _convert(xml_doc, root_node, dict_val)

        return xml_doc

    @classmethod
    def xml_decode(cls, val):

        return xmltodict.parse(cls.utf8(val))

    @classmethod
    def b32_encode(cls, val, standard=False):

        val = cls.utf8(val)

        result = base64.b32encode(val)

        if not standard:
            result = result.rstrip(b'=')

        return cls.basestring(result)

    @classmethod
    def b32_decode(cls, val, standard=False, for_bytes=False):

        val = cls.utf8(val)

        if not standard:

            num = len(val) % 8

            if num > 0:
                val = val + b'=' * (8 - num)

        result = base64.b32decode(val)

        if for_bytes:
            return result
        else:
            return cls.basestring(result)

    @classmethod
    def b64_encode(cls, val, standard=False, for_bytes=False):

        val = cls.utf8(val)

        if standard:

            result = base64.b64encode(val)

        else:

            result = base64.urlsafe_b64encode(val)

            result = result.rstrip(b'=')

        if not for_bytes:
            result = cls.basestring(result)

        return result

    @classmethod
    def b64_decode(cls, val, standard=False, for_bytes=False):

        val = cls.utf8(val)

        if standard:

            result = base64.b64decode(val)

        else:

            num = len(val) % 4

            if num > 0:
                val = val + b'=' * (4 - num)

            result = base64.urlsafe_b64decode(val)

        if for_bytes:
            return result
        else:
            return cls.basestring(result)

    @classmethod
    def jwt_encode(cls, val, key, algorithms=r'HS256'):

        result = jwt.encode(val, key, algorithms)

        return cls.basestring(result)

    @classmethod
    def jwt_decode(cls, val, key, algorithms=None):

        val = cls.utf8(val)

        if algorithms is None:
            algorithms = [r'HS256']

        return jwt.decode(val, key, algorithms)

    @classmethod
    def yaml_encode(cls, data, stream=None):

        return yaml.safe_dump(data, stream)

    @classmethod
    def yaml_decode(cls, stream):

        return yaml.safe_load(stream)

    @classmethod
    def pickle_dumps(cls, val):

        stream = pickle.dumps(val)

        result = zlib.compress(stream)

        return result

    @classmethod
    def pickle_loads(cls, val):

        stream = zlib.decompress(val)

        result = pickle.loads(stream)

        return result

    @classmethod
    def uuid1(cls, node=None, clock_seq=None):

        return uuid.uuid1(node, clock_seq).hex

    @classmethod
    def uuid1_urn(cls, node=None, clock_seq=None):

        return str(uuid.uuid1(node, clock_seq))

    @classmethod
    def crc32(cls, val):

        val = cls.utf8(val)

        return binascii.crc32(val)

    @classmethod
    def md5(cls, val):

        val = cls.utf8(val)

        return hashlib.md5(val).hexdigest()

    @classmethod
    def md5_u32(cls, val):

        val = cls.utf8(val)

        return int(hashlib.md5(val).hexdigest(), 16) >> 96

    @classmethod
    def md5_u64(cls, val):

        val = cls.utf8(val)

        return int(hashlib.md5(val).hexdigest(), 16) >> 64

    @classmethod
    def sha1(cls, val):

        val = cls.utf8(val)

        return hashlib.sha1(val).hexdigest()

    @classmethod
    def sha256(cls, val):

        val = cls.utf8(val)

        return hashlib.sha256(val).hexdigest()

    @classmethod
    def sha512(cls, val):

        val = cls.utf8(val)

        return hashlib.sha512(val).hexdigest()

    @classmethod
    def hmac_md5(cls, key, msg, b64_std=False):

        key = cls.utf8(key)
        msg = cls.utf8(msg)

        result = hmac.new(key, msg, r'md5').digest()

        result = cls.b64_encode(result, b64_std)

        return cls.basestring(result)

    @classmethod
    def hmac_sha1(cls, key, msg, b64_std=False):

        key = cls.utf8(key)
        msg = cls.utf8(msg)

        result = hmac.new(key, msg, r'sha1').digest()

        result = cls.b64_encode(result, b64_std)

        return cls.basestring(result)

    @classmethod
    def ordered_dict(cls, val=None):

        if val is None:
            return OrderedDict()
        else:
            return OrderedDict(sorted(val.items(), key=lambda x: x[0]))

    @classmethod
    def is_iterable(cls, obj, standard=False):

        if standard:
            result = isinstance(obj, Iterable)
        else:
            result = isinstance(obj, (list, tuple,))

        return result

    @classmethod
    def luhn_valid(cls, val):

        return luhn.is_valid(val)

    @classmethod
    def luhn_sign(cls, val):

        result = None

        if val.isdigit():
            result = val + luhn.calc_check_digit(val)

        return result

    @classmethod
    def identity_card(cls, val):

        result = False

        val = val.strip().upper()

        if len(val) == 18:

            weight_factor = (7, 9, 10, 5, 8, 4, 2, 1, 6, 3, 7, 9, 10, 5, 8, 4, 2, 1,)

            verify_list = r'10X98765432'

            check_sum = 0

            for index in range(17):
                check_sum += int(val[index]) * weight_factor[index]

            result = (verify_list[check_sum % 11] == val[17])

        return result

    @classmethod
    def params_join(cls, params, excludes=None):

        if excludes is not None:
            params = {key: val for key, val in params.items() if key not in excludes}

        return r'&'.join(f'{key}={val}' for key, val in sorted(params.items(), key=lambda x: x[0]))

    @classmethod
    def params_sign(cls, *args, **kwargs):

        result = []

        if args:
            result.extend(str(val) for val in args)
            result.sort()

        if kwargs:
            result.append(cls.params_join(kwargs))

        return cls.md5(r'&'.join(result))

    @classmethod
    def get_today_region(cls, today=None):

        if today is None:
            today = cls.today()

        start_date = today
        end_date = today + timedelta(days=1)

        start_time = int(time.mktime(start_date.timetuple()))
        end_time = int(time.mktime(end_date.timetuple())) - 1

        return start_time, end_time

    @classmethod
    def get_month_region(cls, today=None):

        if today is None:
            today = cls.today()

        start_date = today.replace(day=1)

        _, days_in_month = calendar.monthrange(
            start_date.year, start_date.month)

        end_date = start_date + timedelta(days=days_in_month)

        start_time = int(time.mktime(start_date.timetuple()))
        end_time = int(time.mktime(end_date.timetuple())) - 1

        return start_time, end_time

    @classmethod
    def get_week_region(cls, today=None):

        if today is None:
            today = cls.today()

        week_pos = today.weekday()

        start_date = today - timedelta(days=week_pos)
        end_date = today + timedelta(days=(7 - week_pos))

        start_time = int(time.mktime(start_date.timetuple()))
        end_time = int(time.mktime(end_date.timetuple())) - 1

        return start_time, end_time

    @classmethod
    def zip_file(cls, zip_file, *file_paths):

        def _add_to_zip(zf, path, zippath):

            if os.path.isfile(path):

                zf.write(path, zippath, ZIP_DEFLATED)

            elif os.path.isdir(path):

                if zippath:
                    zf.write(path, zippath)

                for nm in os.listdir(path):
                    _add_to_zip(zf, os.path.join(path, nm),
                                os.path.join(zippath, nm))

        with ZipFile(zip_file, r'w') as _zf:

            for path in file_paths:

                zippath = os.path.basename(path)

                if not zippath:
                    zippath = os.path.basename(os.path.dirname(path))
                if zippath in (r'', os.curdir, os.pardir):
                    zippath = r''

                _add_to_zip(_zf, path, zippath)

    @classmethod
    def unzip_file(cls, zip_file, file_paths):

        with ZipFile(zip_file, r'r') as _zf:
            _zf.extractall(file_paths)

    @classmethod
    @cached(cache=TTLCache(maxsize=0xf, ttl=1))
    def get_cpu_percent(cls):

        return psutil.cpu_percent()

    @classmethod
    def kill_process(cls, *args):

        process_names = [val.lower() for val in args]

        for pid in psutil.pids():

            process = psutil.Process(pid)

            if process.name().lower() in process_names:
                process.kill()


class FuncWrapper:
    """函数包装器

    将多个函数包装成一个可调用对象

    """

    def __init__(self):

        self._callables = set()

    def __call__(self, *args, **kwargs):

        for func in self._callables:
            try:
                func(*args, **kwargs)
            except Exception as err:
                Utils.log.exception(str(err))

    @property
    def is_valid(self):

        return len(self._callables) > 0

    def add(self, func):

        if func in self._callables:
            return False
        else:
            self._callables.add(func)
            return True

    def remove(self, func):

        if func in self._callables:
            self._callables.remove(func)
            return True
        else:
            return False



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from abc import ABCMeta, abstractmethod
from asyncio import Queue
from contextlib import asynccontextmanager

from ...extend.base import Utils


class ObjectInterface(metaclass=ABCMeta):

    @abstractmethod
    async def open(self, *args, **kwargs):

        raise NotImplementedError()

    @abstractmethod
    async def close(self, *args, **kwargs):

        raise NotImplementedError()


class ObjectPool(metaclass=ABCMeta):
    """对象池实现
    """

    def __init__(self, maxsize: int):

        self._queue: Queue[ObjectInterface] = Queue(maxsize=maxsize)

    @abstractmethod
    def _create(self) -> ObjectInterface:

        raise NotImplementedError()

    async def open(self, *args, **kwargs):

        while not self._queue.full():
            obj = self._create()
            await obj.open(*args, **kwargs)
            self._queue.put_nowait(obj)

        Utils.log.info(f'ObjectPool {type(self)} initialized: {self._queue.qsize()}')

    async def close(self, *args, **kwargs):

        while not self._queue.empty():
            await self._queue.get_nowait().close(*args, **kwargs)

    @property
    def size(self) -> int:

        return self._queue.maxsize

    @asynccontextmanager
    async def get(self) -> ObjectInterface:

        obj = await self._queue.get()

        try:
            yield obj
        finally:
            await self._queue.put(obj)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import time
import requests

from ...extend.base import Utils
from ...extend.interface import TaskInterface
from ...extend.asyncio.future import Thread


DEFAULT_GROUP_NAME = r'DEFAULT_GROUP'
DEFAULT_PROTOCOL = r'http'
DEFAULT_REQUEST_TIMEOUT = 5
DEFAULT_PULLING_TIMEOUT = 30
DEFAULT_WEIGHT = 1

WORD_SEPARATOR = u'\x02'
LINE_SEPARATOR = u'\x01'


class _NacosInterface(TaskInterface):

    def __init__(self, servers):
        self._task = Thread(target=self._do_task)
        self._servers = Utils.split_str(servers, r',')

    def _do_task(self):
        raise NotImplementedError()

    def start(self):
        self._task.start()

    def stop(self):
        self._task.stop()

    def is_running(self):
        return self._task.is_alive()


class NacosConfig(_NacosInterface):

    def __init__(
            self, listener, servers, data_id, *,
            protocol=None, endpoint=None, group=None, namespace=None,
            request_timeout=None, pulling_timeout=None
    ):

        super().__init__(servers)

        self._listener = listener

        self._protocol = protocol if protocol is not None else DEFAULT_PROTOCOL
        self._endpoint = endpoint if endpoint is not None else r''
        self._namespace = namespace if namespace is not None else r''

        self._data_id = data_id
        self._group = group if group is not None else DEFAULT_GROUP_NAME

        self._content = None
        self._content_hash = None

        self._request_timeout = request_timeout if request_timeout is not None else DEFAULT_REQUEST_TIMEOUT
        self._pulling_timeout = pulling_timeout if pulling_timeout is not None else DEFAULT_PULLING_TIMEOUT

        self._flush()

    def _flush(self):

        self._content, self._content_hash = self.get_config()
        self._listener(self._content)

        Utils.log.info(f'nacos flush config: {self._content_hash}')

    def _do_task(self):

        while True:

            payload = {
                r'Listening-Configs': WORD_SEPARATOR.join(
                    [
                        self._data_id,
                        self._group,
                        self._content_hash,
                        self._namespace,
                    ]
                ) + LINE_SEPARATOR,
            }

            headers = {
                r'Long-Pulling-Timeout': str(self._pulling_timeout * 1000),
            }

            for server in self._servers:

                url = f'{self._protocol}://{server}{self._endpoint}/nacos/v1/cs/configs/listener'

                try:

                    resp = requests.post(
                        url, payload,
                        headers=headers, timeout=self._request_timeout + self._pulling_timeout
                    )

                    if resp.status_code == 200:
                        if resp.text:
                            Utils.log.info(f'nacos config pulling: {resp.text.strip()}')
                            self._flush()
                    else:
                        raise Exception(r'nacos config pulling error')

                except Exception as err:
                    Utils.log.error(str(err))
                    time.sleep(self._request_timeout)
                else:
                    break

            else:

                Utils.log.error(f'nacos config pulling failed: servers unusable')

    def get_config(self):

        content = content_hash = None

        params = {
            r'dataId': self._data_id,
            r'group': self._group,
        }

        if self._namespace:
            params[r'tenant'] = self._namespace

        for server in self._servers:

            url = f'{self._protocol}://{server}{self._endpoint}/nacos/v1/cs/configs'

            try:

                resp = requests.get(url, params, timeout=self._request_timeout)

                if resp.status_code == 200:
                    content = resp.text
                    content_hash = Utils.md5(content)
                else:
                    raise Exception(r'nacos get config error')

            except Exception as err:
                Utils.log.error(str(err))
            else:
                break

        else:

            Utils.log.error(f'nacos get config failed: servers unusable')

        return content, content_hash


class NacosInstanceRegister(_NacosInterface):

    def __init__(
            self, servers, service_name, service_ip, service_port, heartbeat_interval, *,
            protocol=None, endpoint=None, group=None, namespace=None, cluster=None, weight=None,
            request_timeout=None
    ):

        super().__init__(servers)

        self._service_name = service_name
        self._service_ip = service_ip
        self._service_port = service_port

        self._protocol = protocol if protocol is not None else DEFAULT_PROTOCOL
        self._endpoint = endpoint if endpoint is not None else r''
        self._namespace = namespace if namespace is not None else r''

        self._group = group if group is not None else DEFAULT_GROUP_NAME
        self._cluster = cluster
        self._weight = weight if weight is not None else DEFAULT_WEIGHT

        self._request_timeout = request_timeout if request_timeout is not None else DEFAULT_REQUEST_TIMEOUT
        self._heartbeat_interval = heartbeat_interval

    def _do_task(self):

        while True:

            payload = {
                r'serviceName': self._service_name,
                r'namespaceId': self._namespace,
                r'groupName': self._group,
                r'beat': Utils.json_encode(
                    {
                        r'serviceName': self._service_name,
                        r'ip': self._service_ip,
                        r'port': str(self._service_port),
                        r'weight': self._weight,
                        r'ephemeral': True,
                    }
                ),
            }

            for server in self._servers:

                url = f'{self._protocol}://{server}{self._endpoint}/nacos/v1/ns/instance/beat'

                try:

                    resp = requests.put(url, payload, timeout=self._request_timeout)

                    if resp.status_code != 200:
                        raise Exception(r'nacos instance beat error')

                except Exception as err:
                    Utils.log.error(str(err))
                else:
                    break

            else:

                Utils.log.error(f'nacos instance beat failed: servers unusable')

            time.sleep(self._heartbeat_interval)

    def start(self):

        payload = {
            r'serviceName': self._service_name,
            r'ip': self._service_ip,
            r'port': self._service_port,
            r'namespaceId': self._namespace,
            r'weight': self._weight,
            r'enabled': True,
            r'healthy': True,
            r'groupName': self._group,
            r'ephemeral': True,
        }

        if self._cluster:
            payload[r'clusterName'] = self._cluster

        for server in self._servers:

            url = f'{self._protocol}://{server}{self._endpoint}/nacos/v1/ns/instance'

            try:

                resp = requests.post(url, payload, timeout=self._request_timeout)

                if resp.status_code == 200:
                    self._task.start()
                    Utils.log.info(f'nacos instance register: {payload}')
                else:
                    raise Exception(r'nacos instance register error')

            except Exception as err:
                Utils.log.error(str(err))
            else:
                break

        else:

            Utils.log.error(f'nacos instance register failed: servers unusable')


class NacosInstanceQuery(_NacosInterface):

    def __init__(
            self, listener, servers, service_name, listener_interval, *,
            protocol=None, endpoint=None, group=None, namespace=None, cluster=None,
            request_timeout=None
    ):

        super().__init__(servers)

        self._listener = listener

        self._service_name = service_name

        self._protocol = protocol if protocol is not None else DEFAULT_PROTOCOL
        self._endpoint = endpoint if endpoint is not None else r''
        self._namespace = namespace if namespace is not None else r''

        self._group = group if group is not None else DEFAULT_GROUP_NAME
        self._cluster = cluster

        self._content = None
        self._content_hash = None

        self._request_timeout = request_timeout if request_timeout is not None else DEFAULT_REQUEST_TIMEOUT
        self._listener_interval = listener_interval

        self._content, self._content_hash = self._send_query()

    def _do_task(self):

        while True:

            content, content_hash = self._send_query()

            if content_hash != self._content_hash:
                Utils.log.info(f'nacos instance query: {content}')
                self._content, self._content_hash = content, content_hash
                self._listener(self._content)

            time.sleep(self._listener_interval)

    def _send_query(self):

        content = content_hash = None

        params = {
            r'serviceName': self._service_name,
            r'namespaceId': self._namespace,
            r'groupName': self._group,
            r'healthyOnly': True,
        }

        if self._cluster:
            params[r'clusterName'] = self._cluster

        for server in self._servers:

            url = f'{self._protocol}://{server}{self._endpoint}/nacos/v1/ns/instance/list'

            try:

                resp = requests.get(url, params, timeout=self._request_timeout)

                if resp.status_code == 200:
                    content = resp.json()[r'hosts']
                    content_hash = Utils.md5(Utils.json_encode(content))
                else:
                    raise Exception(r'nacos instance query error')

            except Exception as err:
                Utils.log.error(str(err))
            else:
                break

        else:

            Utils.log.error(f'nacos instance query failed: servers unusable')

        return content, content_hash

    def get_host(self):

        if not self._content:
            return None

        return Utils.randhit(self._content, lambda x: int(x[r'weight'] * 100))

    def get_hosts(self):

        return self._content



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import ctypes
import asyncio

from threading import Thread as _Thread
from concurrent.futures import ThreadPoolExecutor

from ..interface import RunnableInterface, TaskInterface

from .base import Utils


class Thread(_Thread):
    """带强制停止功能的线程，如非必要请勿使用
    """

    def stop(self):

        if ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_long(self.ident), ctypes.py_object(SystemExit)
        ) > 1:
            raise SystemError(r'Kill Thread Failed')

    def wait(self, timeout):

        self.join(timeout)

        if self.is_alive():
            self.stop()


class ThreadPool(RunnableInterface):
    """线程池，桥接线程与协程
    """

    def __init__(self, max_workers=None):

        self._executor = ThreadPoolExecutor(max_workers)

    async def run(self, _callable, *args, **kwargs):
        """线程转协程，不支持协程函数
        """

        loop = asyncio.events.get_event_loop()

        if kwargs:

            return await loop.run_in_executor(
                self._executor,
                Utils.func_partial(
                    _callable,
                    *args,
                    **kwargs
                )
            )

        else:

            return await loop.run_in_executor(
                self._executor,
                _callable,
                *args,
            )


class ThreadWorker:
    """通过线程转协程实现普通函数非阻塞的装饰器
    """

    def __init__(self, max_workers=None):

        self._thread_pool = ThreadPool(max_workers)

    def __call__(self, func):

        @Utils.func_wraps(func)
        def _wrapper(*args, **kwargs):
            return self._thread_pool.run(func, *args, **kwargs)

        return _wrapper


class SubProcess(TaskInterface):
    """子进程管理，通过command方式启动子进程
    """

    @classmethod
    async def create(cls, program, *args, stdin=None, stdout=None, stderr=None, **kwargs):

        inst = cls(program, *args, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs)
        await inst.start()

        return inst

    def __init__(self, program, *args, stdin=None, stdout=None, stderr=None, **kwargs):

        self._program = program
        self._args = args
        self._kwargs = kwargs

        self._stdin = asyncio.subprocess.DEVNULL if stdin is None else stdin
        self._stdout = asyncio.subprocess.DEVNULL if stdout is None else stdout
        self._stderr = asyncio.subprocess.DEVNULL if stderr is None else stderr

        self._process = None
        self._process_id = None

    @property
    def pid(self):

        return self._process_id

    @property
    def process(self):

        return self._process

    @property
    def stdin(self):

        return self._process.stdin if self._process is not None else None

    @property
    def stdout(self):

        return self._process.stdout if self._process is not None else None

    @property
    def stderr(self):

        return self._process.stderr if self._process is not None else None

    def create_stdout_task(self):

        return Utils.create_task(self._log_stdout())

    async def _log_stdout(self):

        stream = self._process.stdout

        while self._process.returncode is None:

            content = await stream.readline()

            if content:
                content = content.decode().strip()
                if content:
                    Utils.log.info(content)
            else:
                break

    def create_stderr_task(self):

        return Utils.create_task(self._log_stderr())

    async def _log_stderr(self):

        stream = self._process.stderr

        while self._process.returncode is None:

            content = await stream.readline()

            if content:
                content = content.decode().strip()
                if content:
                    Utils.log.error(content)
            else:
                break

    def is_running(self):

        return self._process is not None and self._process.returncode is None

    async def start(self):

        if self.is_running():
            return False

        self._process = await asyncio.create_subprocess_exec(
            self._program, *self._args,
            stdin=self._stdin,
            stdout=self._stdout,
            stderr=self._stderr,
            **self._kwargs
        )

        self._process_id = self._process.pid

        return True

    async def stop(self):

        if not self.is_running():
            return False

        self._process.kill()
        await self._process.wait()

        return True

    def kill(self):

        if not self.is_running():
            return False

        self._process.kill()

        return True

    async def wait(self, timeout=None):

        if not self.is_running():
            return

        try:
            await asyncio.wait_for(self._process.wait(), timeout=timeout)
        except Exception as err:
            Utils.log.error(err)
        finally:
            await self.stop()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import json
import yaml

from ...extend.config import ConfigureBase

from .client import NacosConfig


class Configure(ConfigureBase):
    """配置类
    """

    __slots__ = [r'_format', r'_client']

    def __init__(
            self, servers, data_id, *,
            protocol=None, endpoint=None, group=None, namespace=None, data_format=r'yaml'
    ):

        super().__init__()

        self._format = data_format

        self._client = NacosConfig(
            self._reload_config, servers, data_id,
            protocol=protocol, endpoint=endpoint, group=group, namespace=namespace
        )

    def _reload_config(self, content):

        self._clear_options()

        if self._format == r'text':
            self._parser.read_string(content)
        elif self._format == r'json':
            self._parser.read_dict(json.loads(content))
        elif self._format == r'yaml':
            self._parser.read_dict(yaml.load(content, yaml.Loader))

        self._load_options()

    def open(self):

        self._client.start()

    def close(self):

        self._client.stop()
        self._clear_options()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys
import signal
import queue
import logging

from abc import ABCMeta, abstractmethod
from datetime import timedelta

from loguru._file_sink import FileSink

from elasticsearch import Elasticsearch, ApiError as ESApiError, NotFoundError as ESNotFoundError, helpers as es_helpers
from confluent_kafka import Producer as KFKProducer

from .trace import get_trace_id
from .base import Utils
from .asyncio.future import Thread


class LogFileRotator:

    @classmethod
    def make(cls, _size=500, _time=r'00:00'):

        return cls(_size, _time).should_rotate

    def __init__(self, _size, _time):

        _size = _size * (1024 ** 2)
        _time = Utils.split_int(_time, r':')

        now_time = Utils.today()

        self._size_limit = _size
        self._time_limit = now_time.replace(hour=_time[0], minute=_time[1])

        if now_time >= self._time_limit:
            self._time_limit += timedelta(days=1)

    def should_rotate(self, message, file):

        file.seek(0, 2)

        if file.tell() + len(message) > self._size_limit:
            return True

        if message.record[r'time'].timestamp() > self._time_limit.timestamp():
            self._time_limit += timedelta(days=1)
            return True

        return False


DEFAULT_LOG_FILE_ROTATOR = LogFileRotator.make()


class InterceptHandler(logging.Handler):
    """日志拦截器
    """

    def emit(self, record):

        Utils.log.opt(
            depth=8,
            exception=record.exc_info
        ).log(
            record.levelname,
            record.getMessage()
        )


class _BaseSink(metaclass=ABCMeta):
    """基础日志投递基类
    """

    def __init__(self, buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5):

        self._buffer = queue.Queue(buffer_maxsize)

        self._bulk_maxsize = bulk_maxsize
        self._write_max_delay = write_max_delay

        self._task = Thread(target=self._do_task)
        self._task.start()

        signal.signal(signal.SIGINT, self.close)
        signal.signal(signal.SIGTERM, self.close)

    def write(self, message):

        try:

            if message.record[r'thread'].id == self._task.ident or not message.record[r'message'].strip():
                return

            log_extra = message.record[r'extra']

            if r'trace_id' not in log_extra:

                trace_id = get_trace_id()

                if trace_id is not None:
                    log_extra[r'trace_id'] = trace_id

            self._buffer.put_nowait(message)

        except queue.Full as _:

            if message.record[r'level'].no > logging.INFO:
                sys.stderr.write(str(message))

        except Exception as err:

            sys.stderr.write(f'{str(err)}\n')

    def close(self, *_):

        self._task.wait(10)

    def _do_task(self):

        messages = []

        while True:

            try:

                for _idx in range(self._write_max_delay):

                    try:

                        messages.append(
                            self._buffer.get(block=True, timeout=1)
                        )

                        self._buffer.task_done()

                    except queue.Empty:

                        pass

                    finally:

                        if len(messages) >= self._bulk_maxsize or (_idx + 1) >= self._write_max_delay:
                            break

                if messages:
                    self._write_logs(messages)
                    messages.clear()

            except Exception as err:

                sys.stderr.write(f'{str(err)}\n')

    @abstractmethod
    def _write_logs(self, logs):

        raise NotImplementedError()


class QueuedFileSink(_BaseSink, FileSink):
    """日志文件队列
    """

    def __init__(
            self, path, *,
            buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5,
            **kwargs
    ):

        _BaseSink.__init__(self, buffer_maxsize, bulk_maxsize, write_max_delay)
        FileSink.__init__(self, path, **kwargs)

    def _write_logs(self, logs):

        for _log in logs:
            FileSink.write(self, _log)


class KafkaSink(_BaseSink):
    """Kafka日志投递
    """

    def __init__(
            self, servers, topic, *,
            buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5,
            **kwargs
    ):

        super().__init__(buffer_maxsize, bulk_maxsize, write_max_delay)

        kwargs[r'bootstrap.servers'] = servers

        self._producer = KFKProducer(kwargs)

        self._topic = topic

    def _write_logs(self, logs):

        for _log in logs:
            self._producer.produce(
                self._topic,
                Utils.json_encode(
                    {
                        r'extra': _log.record[r'extra'],
                        r'process': {
                            r'id': _log.record[r'process'].id,
                            r'name': _log.record[r'process'].name,
                        },
                        r'thread': {
                            r'id': _log.record[r'thread'].id,
                            r'name': _log.record[r'thread'].name,
                        },
                        r'level': {
                            r'no': _log.record[r'level'].no,
                            r'name': _log.record[r'level'].name,
                        },
                        r'module': f"{_log.record[r'name']}:{_log.record[r'function']}:{_log.record[r'line']}",
                        r'message': _log.record[r'message'],
                        r'timestamp': int(_log.record[r'time'].timestamp() * 1000),
                    }
                ),
            )

            self._producer.poll(0)

        self._producer.flush()


class ElasticsearchDataStreamUtil:

    def __init__(
            self, elasticsearch: Elasticsearch, stream_name: str, *,
            rollover_max_age: str = r'1d', rollover_max_primary_shard_size: str = r'50gb', delete_min_age: str = r'30d',
            refresh_interval: str = r'5s', number_of_replicas=0, timestamp_order: str = r'desc'
    ):

        self._elasticsearch = elasticsearch

        self._stream_name = stream_name
        self._policy_name = f'{stream_name}-ilm-policy'

        self._rollover_max_age = rollover_max_age
        self._rollover_max_primary_shard_size = rollover_max_primary_shard_size

        self._delete_min_age = delete_min_age
        self._refresh_interval = refresh_interval
        self._number_of_replicas = number_of_replicas
        self._timestamp_order = timestamp_order

    def initialize(self):

        try:
            self._elasticsearch.indices.get_data_stream(name=self._stream_name)
        except ESNotFoundError as _:
            self._create_lifecycle()
            self._create_index_template()
        except ESApiError as err:
            sys.stderr.write(f'{str(err.info)}\n')
        except Exception as err:
            sys.stderr.write(f'{str(err)}\n')

    def _create_lifecycle(self):

        try:

            policy = {
                r'phases': {
                    r'hot': {
                        r'actions': {
                            r'rollover': {
                                r'max_age': self._rollover_max_age,
                                r'max_primary_shard_size': self._rollover_max_primary_shard_size,
                            },
                            r'set_priority': {
                                r'priority': 100,
                            }
                        },
                        r'min_age': r'0ms',
                    },
                    r'delete': {
                        r'actions': {
                            r'delete': {}
                        },
                        r'min_age': self._delete_min_age,
                    },
                },
            }

            self._elasticsearch.ilm.put_lifecycle(name=self._policy_name, policy=policy)

        except ESApiError as err:

            sys.stderr.write(f'{str(err.info)}\n')

        except Exception as err:

            sys.stderr.write(f'{str(err)}\n')

    def _create_index_template(self):

        try:

            mappings = {
                r'dynamic': r'strict',
                r'properties': {
                    r'extra': {
                        r'type': r'flattened',
                    },
                    r'process': {
                        r'properties': {
                            r'id': {
                                r'type': r'keyword',
                            },
                            r'name': {
                                r'type': r'keyword',
                                r'ignore_above': 64,
                            },
                        },
                    },
                    r'thread': {
                        r'properties': {
                            r'id': {
                                r'type': r'keyword',
                            },
                            r'name': {
                                r'type': r'keyword',
                                r'ignore_above': 64,
                            },
                        },
                    },
                    r'level': {
                        r'properties': {
                            r'no': {
                                r'type': r'integer',
                            },
                            r'name': {
                                r'type': r'keyword',
                                r'ignore_above': 64,
                            },
                        },
                    },
                    r'module': {
                        r'type': r'text',
                        r'norms': False,
                    },
                    r'message': {
                        r'type': r'text',
                        r'norms': False,
                    },
                    r'@timestamp': {
                        r'type': r'date',
                    },
                }
            }

            template = {
                r'settings': {
                    r'index': {
                        r'lifecycle': {
                            r'name': self._policy_name,
                        },
                        r'refresh_interval': self._refresh_interval,
                        r'number_of_replicas': self._number_of_replicas,
                        r'sort': {
                            r'field': r'@timestamp',
                            r'order': self._timestamp_order,
                        }
                    },
                },
                r'mappings': mappings,
            }

            self._elasticsearch.indices.put_index_template(
                name=self._stream_name,
                template=template,
                index_patterns=[f'{self._stream_name}*'],
                data_stream={},
            )

        except ESApiError as err:

            sys.stderr.write(f'{str(err.info)}\n')

        except Exception as err:

            sys.stderr.write(f'{str(err)}\n')


class ElasticsearchSink(_BaseSink):
    """Elasticsearch日志投递
    """

    def __init__(
            self, hosts, index, *,
            buffer_maxsize: int = 0xffff, bulk_maxsize: int = 0x800, write_max_delay: int = 5,
            rollover_max_age: str = r'1d', rollover_max_primary_shard_size: str = r'50gb', delete_min_age: str = r'30d',
            refresh_interval: str = r'5s', number_of_replicas=0, timestamp_order: str = r'desc',
            **kwargs):

        super().__init__(buffer_maxsize, bulk_maxsize, write_max_delay)

        self._elasticsearch = Elasticsearch(hosts, **kwargs)

        ElasticsearchDataStreamUtil(
            self._elasticsearch, index,
            rollover_max_age=rollover_max_age,
            rollover_max_primary_shard_size=rollover_max_primary_shard_size,
            delete_min_age=delete_min_age,
            refresh_interval=refresh_interval,
            number_of_replicas=number_of_replicas,
            timestamp_order=timestamp_order
        ).initialize()

        self._index = index

    def _write_logs(self, logs):

        es_helpers.bulk(
            self._elasticsearch,
            actions=[
                {
                    r'_op_type': r'create',
                    r'_index': self._index,
                    r'extra': _log.record[r'extra'],
                    r'process': {
                        r'id': _log.record[r'process'].id,
                        r'name': _log.record[r'process'].name,
                    },
                    r'thread': {
                        r'id': _log.record[r'thread'].id,
                        r'name': _log.record[r'thread'].name,
                    },
                    r'level': {
                        r'no': _log.record[r'level'].no,
                        r'name': _log.record[r'level'].name,
                    },
                    r'module': f"{_log.record[r'name']}:{_log.record[r'function']}:{_log.record[r'line']}",
                    r'message': _log.record[r'message'],
                    r'@timestamp': int(_log.record[r'time'].timestamp() * 1000),
                }
                for _log in logs
            ]
        )


DEFAULT_LOG_FILE_NAME = r'runtime_{time}.log'


def init_logger(
        level, *, handler=None,
        file_path=None, file_name=DEFAULT_LOG_FILE_NAME,
        file_rotation=DEFAULT_LOG_FILE_ROTATOR, file_retention=0xff,
        extra=None, enqueue=False, debug=False
):

    level = level.upper()

    Utils.log.remove()

    if extra is not None:

        extra = {_key: _val for _key, _val in extra.items() if _val is not None}

        if extra:
            Utils.log.configure(extra=extra)

    if handler or file_path:

        if handler:
            Utils.log.add(
                handler,
                level=level,
                enqueue=enqueue,
                backtrace=debug
            )

        if file_path:

            _file_name, _file_ext_name = os.path.splitext(file_name)

            Utils.log.add(
                QueuedFileSink(
                    Utils.path.join(file_path, _file_name + '.pid-' + str(Utils.getpid()) + _file_ext_name),
                    rotation=file_rotation,
                    retention=file_retention
                ),
                level=level,
                enqueue=enqueue,
                backtrace=debug
            )

    else:

        Utils.log.add(
            sys.stderr,
            level=level,
            enqueue=enqueue,
            backtrace=debug
        )

    logger = logging.getLogger()
    logger.setLevel(level)
    logger.addHandler(InterceptHandler())



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from contextlib import contextmanager

from .base import Utils


# 基础异常
class BaseError(Exception):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)


class Ignore(BaseError):
    """可忽略的异常

    用于with语句块跳出，或者需要跳出多层逻辑的情况

    """

    def __init__(self, *args, layers=1, **kwargs):

        super().__init__(*args, **kwargs)

        self._layers = layers

        info = str(self)

        if info:
            Utils.log.warning(info)

    def throw(self):

        if self._layers > 0:
            self._layers -= 1

        return self._layers != 0


@contextmanager
def catch_warning():
    """异常捕获，打印warning级日志

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    try:
        yield
    except Ignore as err:
        if err.throw():
            raise err
    except Exception as err:
        Utils.log.warning(str(err))


@contextmanager
def catch_error():
    """异常捕获，打印error级日志

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    try:
        yield
    except Ignore as err:
        if err.throw():
            raise err
    except Exception as err:
        Utils.log.exception(str(err))


# 数据库只读限制异常
class MySQLReadOnlyError(BaseError):
    pass


# NTP校准异常
class NTPCalibrateError(BaseError):
    pass


# Router异常
class RouterError(BaseError):
    pass



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import yarl
import logging
import asyncio
import aio_pika


aio_pika.logger.setLevel(logging.INFO)


def create_connection(
        url: yarl.URL, timeout: aio_pika.abc.TimeoutType = None, **kwargs
) -> aio_pika.RobustConnection:

    connection: aio_pika.RobustConnection = aio_pika.RobustConnection(url, **kwargs)

    asyncio.create_task(connection.connect(timeout))

    return connection
